FORM_CONTENT_TYPES = [
    'application/x-www-form-urlencoded',
    'multipart/form-data'
]

METHODS = set([
    "get",
    "put",
    "post",
    "delete",
    "options",
    "head",
    "patch",
    "trace"
])
